/// QasTools: Desktop toolset for the Linux sound system ALSA.
/// \copyright See COPYING file.

#pragma once

#include "qsnd/alsa.hpp"
#include "qsnd/event_types.hpp"
#include <QObject>
#include <QSocketNotifier>
#include <array>
#include <memory>
#include <vector>

// Forward declaration
namespace QSnd
{
class Mixer_HCTL_Info_Db;
class Mixer_HCTL_Elem;
class Mixer_HCTL_Elem_Group;
} // namespace QSnd

namespace QSnd
{

/// @brief Mixer_HCTL
///
/// Brings Qt and ALSA objects together but without
/// any GUI objects
///
class Mixer_HCTL : public QObject
{
  Q_OBJECT

  // Public methods
  public:
  // -- Construction

  Mixer_HCTL ( QObject * parent_n = nullptr );

  ~Mixer_HCTL ();

  // -- Accessors

  const Mixer_HCTL_Info_Db *
  info_db () const
  {
    return _info_db.get ();
  }

  snd_hctl_t *
  snd_hctl ()
  {
    return _snd_hctl;
  }

  // -- Init / close

  /// @return On an error a negative error code
  int
  open ( const QString & dev_name_n );

  void
  close ();

  bool
  is_open () const
  {
    return ( _snd_hctl != nullptr );
  }

  // -- Error strings / codes

  const QString &
  err_func () const
  {
    return _err_func;
  }

  const QString &
  err_message () const
  {
    return _err_message;
  }

  // -- Elements

  ::std::size_t
  num_elems () const
  {
    return _mixer_elems.size ();
  }

  const Mixer_HCTL_Elem *
  elem ( ::std::size_t idx_n ) const
  {
    return _mixer_elems[ idx_n ].get ();
  }

  Mixer_HCTL_Elem *
  elem ( ::std::size_t idx_n )
  {
    return _mixer_elems[ idx_n ].get ();
  }

  // -- Interface types

  unsigned int
  iface_type_count ( ::std::size_t type_idx_n ) const
  {
    return _iface_type_count[ type_idx_n ];
  }

  unsigned int
  iface_types_avail () const
  {
    return _iface_num_types;
  }

  unsigned int
  iface_avail_type ( ::std::size_t type_idx_n ) const
  {
    return _iface_avail_types[ type_idx_n ];
  }

  // -- Alsa callbacks

  static int
  alsa_callback_hctl ( snd_hctl_t * snd_hctl_n,
                       unsigned int mask_n,
                       snd_hctl_elem_t * elem_n );

  // -- Signals

  Q_SIGNAL
  void
  sig_mixer_reload_request ();

  // -- Public slots

  Q_SLOT
  void
  request_reload ();

  // -- Protected slots
  protected:
  Q_SLOT
  void
  socket_event ( int socket_id );

  // Protected methods
  protected:
  int
  create_mixer_elems ();

  // Socket notifiers

  int
  create_socket_notifiers ();

  void
  set_socked_notifiers_enabled ( bool flag_n );

  void
  signalize_all_changes ();

  bool
  event ( QEvent * event_n ) override;

  // Private attributes
  private:
  ::std::unique_ptr< ::QSnd::Mixer_HCTL_Info_Db > _info_db;
  snd_hctl_t * _snd_hctl = nullptr;

  ::std::vector< ::std::unique_ptr< Mixer_HCTL_Elem > > _mixer_elems;
  ::std::vector< pollfd > _pollfds;
  ::std::vector< ::std::unique_ptr< QSocketNotifier > > _socket_notifiers;

  unsigned int _iface_num_types = 0;
  ::std::array< unsigned int, 7 > _iface_avail_types;
  ::std::array< unsigned int, 7 > _iface_type_count;

  bool _update_requested = false;
  bool _reload_requested = false;

  QString _err_func;
  QString _err_message;
};

} // namespace QSnd
